import os
import sys
import defaults
import yaml
import errno
import shutil
import messages as messages


class NewProject(object):
    
    config_path = "wrangler.yaml"
    directories = ["content", "templates", "var", "www", "lib"]
    default_files = [
            {
                "path": "content/index.md",
                "content": defaults.content_template()
            },
            {
                "path": "templates/template.j2",
                "content": defaults.template_template()
            },
            {
                "path": "wrangler.yaml",
                "content": defaults.config_template()
            },
            {
                "path": "lib/examples.py",
                "content": defaults.lib_template()
            }
        ]

    def __init__(self, path, reporter):
        self.path = path
        self.abspath = os.path.abspath(self.path)
        self.reporter = reporter
        self.reporter.log(messages.create_project % (self.abspath), "green")
        self.create_directory_structure()
        self.create_default_files()
        return None

    def ensure_path_exists(self, path):
        try:
            os.makedirs(path)
            self.reporter.log("Creating '%s'" % (path), "green")
        except OSError as exception:
            if exception.errno != errno.EEXIST:
                raise
            else:
                self.reporter.log(messages.already_exists % (path), "red")

    def create_default_files(self):
        for f in self.default_files:
            safe = self.safe_path(f["path"])
            if not os.path.exists(safe):
                shutil.copy2(f["content"], f["path"])
                self.reporter.log(messages.file_created % (safe), "green")
            else:
                self.reporter.log(messages.already_exists % (safe), "red")


    def create_directory_structure(self):
        for path in self.directories:
            safe = self.safe_path(path)
            self.ensure_path_exists(safe)

    def write_file(self, path, content):
        with open(path, 'w') as outfile:
            outfile.write(content)

    def safe_path(self, path):
        return os.path.join(self.abspath, path)




import os
import yaml

cwd = os.path.dirname(os.path.abspath(__file__))
content = "defaults/index.md"
template = "defaults/template.j2"
config = "defaults/wrangler.yaml"
lib = "defaults/examples.py"

def load_defaults():
    return yaml.load(file(config_template(), 'r'))

def content_template():
    return "%s/%s" % (cwd, content)

def template_template():
    return "%s/%s" % (cwd, template)

def config_template():
    return "%s/%s" % (cwd, config)

def lib_template():
    return "%s/%s" % (cwd, lib)


import os
import subprocess
import shelve 
import sys 
import Core
import Parsers
from blinker import signal
import messages as messages

# Takes a directory and transforms the matching elements into Page instances.
class Reader():
    default_class = "Page"
    parsers = {}
    classes = {
        default_class: Core.Page
    }

    def __init__(self, config, reporter):
        conf = config["wrangler"]
        self.input_dir = conf['input_dir']
        self.ignore_files = conf['ignore']
        self.nocache = conf['nocache']
        self.lib_path = conf["lib_path"]
        self.config = config
        self.reporter = reporter
        formats_allowed_by_config = conf['data_formats'] if 'data_formats' in conf else None

        self.check_custom_default_class()
        self.data_formats = self.register_parsers(formats_allowed_by_config)


    def register_parsers(self, formats_allowed_by_config=None):
        formats = []

        if formats_allowed_by_config == []:
            exit("You need at least one file format configured. Check your wrangler.yaml file.")

        for cls in Core.Parser.__subclasses__():
            
            if hasattr(cls, "accepts"):
                acceptable_types = []
                parser = cls(self.input_dir, "")
                for file_format in cls.accepts:
                    if formats_allowed_by_config:
                        if file_format in formats_allowed_by_config:
                            acceptable_types.append(file_format)
                            self.parsers[file_format] =  parser
                    else:
                        acceptable_types.append(file_format)
                        self.parsers[file_format] =  parser

                self.reporter.verbose("Created a %s to handle %s" % (cls.__name__, acceptable_types), "blue")
                formats += acceptable_types

        return formats


    def check_custom_default_class(self):
        for cls in Core.Page.__subclasses__():
            if cls.__name__ == self.default_class:
                self.classes[self.default_class] = cls
                self.reporter.verbose("Set default class to new default subclass '%s'" % (cls.__name__), "blue")

           

    def load_class(self, file_class):
        if not file_class in self.classes:
            for cls in Core.Page.__subclasses__():
                if cls.__name__ == file_class:
                    self.classes[file_class] = cls
                    self.reporter.verbose("Loaded class %s" % (file_class), "blue")

        return self.classes[file_class]


    def init_cache(self):
        if (self.nocache):
            try:
                os.remove(self.config['wrangler']['build_cache_file'])
            except OSError:
                pass
        try:
            return shelve.open(self.config['wrangler']['build_cache_file'])
        except:
            self.reporter.log("Couldn't open cache at %s. Try running `wrangler clean`" % (self.config['wrangler']['build_cache_file']), "red")
            exit() 
    
    def save_cache(self, shelf):
        try:
            shelf.sync()
        finally:
            shelf.close()


    def dir_as_tree(self, path, parent=None):
        """Recursive function that walks a directory and returns a tree
        of nested nodes.
        """

        basename = os.path.basename(path)
        node = Core.Node(basename, path, parent)
        self.graph.add(node)

        # Gather some more information on this path here
        # and write it to attributes
        if os.path.isdir(path):
            # Recurse
            node.tag = 'dir'
            for item in sorted(os.listdir(path)):
                item_path = os.path.join(path, item)

                # Ignore all the crufty things that get scraped, as well as hidden fies
                if not os.path.basename(item_path).startswith(".") and not ":" in item_path and not "%" in item_path and not "?" in item_path:
                    child_node = self.dir_as_tree(item_path, node)
                    node.add_child(child_node)
            return node
        else:
            node.tag = 'file'

            if basename.startswith("index") and basename.endswith(tuple(self.data_formats)):
                node.is_index = True 
            return node


    def comparitor(self, a, b):
        a_cargo = a.get_cargo()
        b_cargo = b.get_cargo()

        if not a_cargo:
            if a.tag == "dir":
                for _child in a.children:
                    filename = os.path.basename(_child.path)

                    if filename.startswith("index") and filename.endswith(tuple(self.data_formats)):
                        a_cargo = _child.get_cargo()
               
        if not b_cargo:
            if b.tag == "dir":
                for _child in b.children:
                    filename = os.path.basename(_child.path)
                    if filename.startswith("index") and filename.endswith(tuple(self.data_formats)):
                        b_cargo = _child.get_cargo()
                       
        try:
            a_weight = a_cargo.get_weight()
        except:
            a_weight = 0

        try:
            b_weight = b_cargo.get_weight()
        except:
            b_weight = 0

        if a_weight < b_weight: return -1
        if a_weight > b_weight: return 1

        return 0


    def recursive_sort(self, node):
        if node.tag == 'dir':
            node.children = sorted(node.children, cmp=self.comparitor)
            for n in node.children:
                self.recursive_sort(n)
        return node


    def fetch(self):
        shelf = self.init_cache()

        self.reporter.log(messages.start_render % (self.data_formats), "blue")

        self.graph = Core.NodeGraph()
        root_node = self.dir_as_tree(self.input_dir)
        self.graph.root(root_node)
        
        for key, node in self.graph.all().items():
            if node.tag == "file":
                node.add_cargo(self.new_item(shelf, node.path))
        
        self.recursive_sort(root_node)
        self.save_cache(shelf)

        return self.graph


    def load_parser_by_format(self, data_format):
        try: 
            return self.parsers[data_format]
        except:
            self.reporter.verbose("No parser found for %s" % (data_format), "red")


    def new_item(self, shelf, filename):

        file_format = os.path.splitext(filename)[1].replace(os.extsep, "")
        parser = self.load_parser_by_format(file_format)

        try:
            mtime = os.path.getmtime(filename)
        except:
            print "Oops! Couldn't load \"%s\". Check the path exists." % (filename)
            return False

        if shelf.has_key(filename):
            try:
                last_modified = shelf[filename].get_mtime()
            except:
                self.reporter.log(messages.unpickle_error % (filename), "red")
                exit()


        if (not shelf.has_key(filename)) or (last_modified < mtime) or (self.nocache):
            
            # Check if custom class is set, otherwise make it a page... 
            if parser:
                page_data = parser.load(filename);

                if page_data:
                    page_classname = self.default_class

                    if "class" in page_data["meta"] and page_data["meta"]["class"] != None:
                        page_classname = page_data["meta"]["class"]

                    PageClass = self.load_class(page_classname)
                    new_page = PageClass(page_data, self.config)
                    shelf[filename] = new_page
                    print "\033[1;35mCaching \033[0m\033[2m%s\033[0m" % (filename)
                    return new_page
        else:
            if filename in shelf:
                return shelf[filename]



import os
import markdown as md
import sys
import traceback
import glob
import Core
import utilities as util
import types

from docutils.core import publish_parts
from jinja2 import Environment, FileSystemLoader, meta, BaseLoader, TemplateNotFound, ChoiceLoader, Template, Markup
from SilentUndefined import SilentUndefined
from MarkdownExtension import MarkdownExtension
from blinker import signal

def rst_filter(s):
    return Markup(publish_parts(source=s, writer_name='html')['body'])

def markdown_filter(value):
    """
    Jinja2 markdown filter with support for extensions.
    """
    output = value
    marked = md.Markdown(
        extensions=['extra','meta','sane_lists'],
        output_format='html5',
        safe_mode=False 
        )
    out = marked.convert(output)
    return out




class JinjaStaticRenderer(Core.Renderer):

    def init(self):

        # Setup the env
        self.env = Environment(
            undefined=SilentUndefined,
            loader=FileSystemLoader(self.config['templates_dir']),
            extensions=[
                MarkdownExtension
            ],
            trim_blocks = True,
            lstrip_blocks = True
            )
        
        self.env.filters["markdown"] = markdown_filter
        self.env.filters['rst'] = rst_filter

        if not os.path.exists(self.config['templates_dir']):
            self.reporter.log("Couldn't find %s. Check your templates exist, and your config in wrangler.yaml" % (self.config['templates_dir']), "red")
            exit()

        # Load up some custom, project specific filters
        # if "lib_path" in self.config and os.path.exists(self.config["lib_path"]):

        sig = signal("template_filter")
        results = sig.send("renderer", env=self.env, config=self.config)

        for filter_function in results:
            name = filter_function[1].__name__
            self.env.filters[name] = filter_function[1]


        self.template_trees = {}
        self.template_modified_times = {}
        
        var_path = os.path.dirname(self.config['compiled_templates_file'])

        self.reporter.verbose("Ensure directory exists: \033[34m%s\033[0m" % (var_path))
        util.ensure_dir(var_path)

        self.reporter.verbose("Loading templates from: \033[34m%s\033[0m" % (self.config['templates_dir']))
        self.env.compile_templates(
            self.config['compiled_templates_file'],
            ignore_errors=False,
            filter_func=self.filter_hidden_files 
            )

        self.reporter.verbose("Compile templates to .zip: \033[32m%s\033[0m" % (self.config['compiled_templates_file']))
    
    def filter_hidden_files(self, filename):
        _name = os.path.basename(filename)
        _path = os.path.dirname(filename)

        for _segment in _path.split(os.sep):
            if _segment.startswith("."):
                return False
        if _name.startswith("."):
            return False;
        
        self.reporter.verbose("Loading template: %s" % (_name))

        return True;


    def get_template_mtime(self, template):
        """
        Request the template's mtime. If we don't have it yet, cache it.
        """
        if not template in self.template_modified_times:
            self.template_modified_times[template] = os.path.getmtime("%s/%s" % (self.config['templates_dir'], template) )
        return self.template_modified_times[template]


    def get_referenced_templates(self, template):
        """
        Request the template's Abstract Syntax Tree so we can find other template references.
        Store the reference list in a dictionary
        """

        if not template in self.template_trees:
            try: 
                ast = self.env.parse(self.env.loader.get_source(self.env, template))
                self.template_trees[template] = list(meta.find_referenced_templates(ast))
            except:
                self.template_trees[template] = list()

            self.reporter.verbose("\033[34m%s\033[0m references templates \033[32m%s\033[0m" % (template, self.template_trees[template]))

        return self.template_trees[template]


    def render(self, item, **kwargs):
        template_name = item.get_template()
        template = self.load_template(template_name)

        if template and (self.should_render_item(item, template_name)):
            sig = signal("wranglerRenderItem")
            result = sig.send('item', item=item, template_name=template_name)
            return (template.render(kwargs), item)

        return (False, item)


    def load_template(self, template_name):
        template_object = False
        try:
            template_object = self.env.get_template(template_name)
        except:
            print "\033[1;91mCouldn't parse `%s`. Check the template for errors \033[0;37m" % (template)
            
            if self.config["verbose"] == True:
                traceback.print_exc()
            print "\033[0m"
            return False
        finally:
            return template_object


    def should_render_item(self, item, template_name):
        force_render = 0
        last_build_time = self.reporter.get_last_build_time()
        output_path = item.get_output_path()

        if last_build_time == 0:
            return True

        config_modified_time = self.reporter.get_config_modified_time()
        
        if last_build_time == 0 or config_modified_time >= last_build_time:
            return True

        if self.config["force"] == True:
            return True

        if output_path and not os.path.exists(output_path):
            return True
        
        if item.get_modified_time() >= last_build_time:
            return True

        template_modified_time = 0

        if template_name:

            referenced_templates = self.get_referenced_templates(template_name)
            template_modified_time = self.get_template_mtime(template_name)
            
            if template_modified_time >= last_build_time:
                return True

            ref_template_mtime = 0

            for t in referenced_templates:
                time = self.get_template_mtime(t)
                if time >= last_build_time:
                    ref_template_mtime = time
                    return True

        return False
